function [param]=sigm_fit(x,y,fixed_params,initial_params,plot_flag)
% Optimization of parameters of the sigmoid function
%
% Syntax:
%       [param]=sigm_fit(x,y,mse_max)       
%
%       that is the same that
%       [param]=sigm_fit(x,y,mse_max,[],[],[])     % no fixed_params, automatic initial_params
%
%       [param]=sigm_fit(x,y,mse_max,fixed_params)        % automatic initial_params
%       [param]=sigm_fit(x,y,mse_max,[],initial_params)   % use it when the estimation is poor
%       [param]=sigm_fit(x,y,mse_max,fixed_params,initial_params,plot_flag)
%
% param = [min, max, x50, slope]
%
% if fixed_params=[NaN, NaN , NaN , NaN]        % or fixed_params=[]
% optimization of "min", "max", "x50" and "slope" (default)
%
% if fixed_params=[0, 1 , NaN , NaN]
% optimization of x50 and slope of a sigmoid of ranging from 0 to 1
%
%
% Additional information in the second output, STAT
% [param,stat]=sigm_fit(x,y,fixed_params,initial_params,plot_flag)
%
%
% Example:
% %% generate data vectors (x and y)
% fsigm = @(param,xval) param(1)+(param(2)-param(1))./(1+10.^((param(3)-xval)*param(4)))
% param=[0 1 5 1];  % "min", "max", "x50", "slope"
% x=0:0.1:10;
% y=fsigm(param,x) + 0.1*randn(size(x));
%
% %% standard parameter estimation
% [estimated_params]=sigm_fit(x,y,mse_max)
%
% %% parameter estimation with forced 0.5 fixed min
% [estimated_params]=sigm_fit(x,y,[0.5 NaN NaN NaN])
%
% %% parameter estimation without plotting
% [estimated_params]=sigm_fit(x,y,[],[],0)
%
%
% Doubts, bugs: rpavao@gmail.com
% Downloaded from http://www.mathworks.com/matlabcentral/fileexchange/42641-sigmoid-logistic-curve-fit

% Addition of sorting of y vector first and doing an initial fit on on it to improve final fit, added by
% arthur.avramiea@gmail.com

x=x(:);
y=y(:);

%giving the initial parameters based on a sorted vector y gives better
%final estimates in cases where y is noisy.
y_sorted = sort(y);

if nargin<=1 %fail
    fprintf('');
    help sigm_fit
    return
end

automatic_initial_params=[quantile(y_sorted,0.05) quantile(y_sorted,0.95) NaN 1];
if sum(y_sorted==quantile(y_sorted,0.5))==0
    temp=x(y_sorted==quantile(y(2:end),0.5));    
else
    temp=x(y_sorted==quantile(y,0.5));
end
automatic_initial_params(3)=temp(1);

if nargin==2 %simplest valid input
    fixed_params=NaN(1,4);
    initial_params=automatic_initial_params;
    plot_flag=0;    
end
if nargin==3
    initial_params=automatic_initial_params;
    plot_flag=0;    
end
if nargin==4
    plot_flag=0;    
end

if exist('fixed_params','var')
    if isempty(fixed_params)
        fixed_params=NaN(1,4);
    end
end
if exist('initial_params','var')
    if isempty(initial_params)
        initial_params=automatic_initial_params;
    end
end
if exist('plot_flag','var')
    if isempty(plot_flag)
        plot_flag=1;
    end
end

f_str='f = @(param,xval)';
free_param_count=0;
bool_vec=NaN(1,4);
for i=1:4;
    if isnan(fixed_params(i))
        free_param_count=free_param_count+1;
        f_str=[f_str ' param(' num2str(free_param_count) ')'];
        bool_vec(i)=1;
    else
        f_str=[f_str ' ' num2str(fixed_params(i))];
        bool_vec(i)=0;
    end
    if i==1; f_str=[f_str ' + (']; end
    if i==2;
        if isnan(fixed_params(1))            
            f_str=[f_str '-param(1) )./ (   1 + 10.^( (']; 
        else
            f_str=[f_str '-' num2str(fixed_params(1)) ')./ (1 + 10.^((']; 
        end
    end    
    if i==3; f_str=[f_str ' - xval ) *']; end
    if i==4; f_str=[f_str ' )   );']; end
end

eval(f_str)

%first fit on sorted data
[params_sorted_data,RESID,J,COVB,MSE] = nlinfit(x,sort(y),f,initial_params(bool_vec==1));

%then use these as starting parameters to fit the actual data
[params_actual_data,RESID,J,COVB,MSE] = nlinfit(x,y,f,params_sorted_data);

param=params_actual_data';
%     
if plot_flag==1
    x_vector=min(x):(max(x)-min(x))/100:max(x);
    plot(x,y,'k.',x_vector,f(param(isnan(fixed_params)),x_vector),'r-')
    xlim([min(x) max(x)])
end
