#     Originally created by Arthur-Ervin Avramiea (2020), arthur.avramiea@gmail.com

#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
# 
#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU General Public License for more details.
# 
#     You should have received a copy of the GNU General Public License
#     along with this program.  If not, see <http://www.gnu.org/licenses/>.

import shutil
import platform

from brian2 import *

from utils import output_utils
from utils import parameter_utils as p
from utils import network_utils
from utils import analysis_utils
from configuration import configuration as conf
import os
import numpy_indexed as npi
import sys

########################################## BRIAN SETTINGS
prefs.codegen.target = 'cython'
if "Windows" not in platform.platform():
    prefs.codegen.cpp.extra_compile_args_gcc = ['-w', '-Ofast', '-march=native']

########################################## SIMULATION PARAMETERS - LOAD AND PREPROCESS
configuration = conf.get_configuration()
simulation_param_files_path = configuration[conf.SIMULATION_PARAM_FILES_PATH]

parameters = p.get_default_parameters()
#set random seed
random_seed = p.get_random_seed()
parameters[p.RANDOM_SEED]=random_seed
parameters[p.RUN_COMMAND]='python '+' '.join(sys.argv)

#override with params passed as arguments
params_file_path = os.path.join(simulation_param_files_path, sys.argv[1])

rest_of_arguments = sys.argv[2:]
params_to_override = p.load_params_file(params_file_path,rest_of_arguments)

for crt_key in params_to_override.keys():
    parameters[crt_key]=params_to_override[crt_key]

# one way to load the network is by specifying the network format path. this is especially useful when doing running phase space
# another way is just specify the path directly to that network structure as a parameter

#here we are taking care of the case where we specify the network format path. 
if p.NETWORK_TO_LOAD_PATH_FORMAT in parameters.keys():
    network_path_parameters_keys = parameters[p.NETWORK_TO_LOAD_PATH_INPUT_PARAMETERS].split(",")
    network_path_parameters_values = []
    for parameter_key in network_path_parameters_keys:
        network_path_parameters_values.append(parameters[parameter_key.strip()])
    
    parameters[p.NETWORK_TO_LOAD_PATH]=(configuration[conf.DATA_FILE_PATH] + parameters[p.NETWORK_TO_LOAD_PATH_FORMAT]+'/SimulationData.h5') \
                  % tuple(network_path_parameters_values)

#initialize output
[output_dir_path,tmp_dir_path] = output_utils.create_simulation_directories(parameters)

print("running simulation at:")
print(output_dir_path)

########################################## CREATE OR LOAD NETWORK

num_neurons=int(parameters[p.NUMBER_NEURONS])
grid_size=parameters[p.GRID_SIZE]

num_excitatory = int(round(parameters[p.FRACTION_EXCITATORY]*num_neurons))
num_inhibitory = num_neurons-num_excitatory

if not p.NETWORK_TO_LOAD_PATH in parameters.keys():
    all_indexes = np.array(range(0,num_neurons))
    
    #determine placement of inhibitory neurons on the grid 
    
    #first all neurons are set to excitatory
    types=np.zeros(num_neurons)+1
    
    #get random positions for inhibitory neurons (using mitchell's best candidate algorithm 
    #to ensure a more uniform spread of neurons
    [xx,yy] = network_utils.get_random_points_placement(grid_size,num_inhibitory)
    indices_inhibitory = xx*grid_size+yy
    types[indices_inhibitory]=0
    
    from_to = network_utils.get_network(num_neurons,types,parameters)
        
    wEE = parameters[p.W_EX_EX]
    wEI = parameters[p.W_EX_IN]
    wIE = parameters[p.W_IN_EX]
    wII = parameters[p.W_IN_IN]
    
    weights = np.zeros(np.shape(from_to)[0])
    weights[ np.where( (types[from_to[:,0]]==1)  & (types[from_to[:,1]]==1) )[0]]=wEE
    weights[ np.where( (types[from_to[:,0]]==1)  & (types[from_to[:,1]]==0) )[0]]=wEI
    weights[ np.where( (types[from_to[:,0]]==0)  & (types[from_to[:,1]]==1) )[0]]=wIE
    weights[ np.where( (types[from_to[:,0]]==0)  & (types[from_to[:,1]]==0) )[0]]=wII
else:
    [types,from_to,weights]=network_utils.get_network_arrays_from_path(parameters[p.NETWORK_TO_LOAD_PATH])
    
    wEE = weights[ np.where( (types[from_to[:,0]]==1)  & (types[from_to[:,1]]==1) )[0]]
    wEI = weights[ np.where( (types[from_to[:,0]]==1)  & (types[from_to[:,1]]==0) )[0]]
    wIE = weights[ np.where( (types[from_to[:,0]]==0)  & (types[from_to[:,1]]==1) )[0]]
    wII = weights[ np.where( (types[from_to[:,0]]==0)  & (types[from_to[:,1]]==0) )[0]]

# Preprocessing of network indexes to be able to separate between excitatory and inhibitory neuron groups

excitatory_indexes = np.where(types==1)[0]
inhibitory_indexes = np.where(types==0)[0]

excitatory_indexes_sorted = np.sort(excitatory_indexes)
inhibitory_indexes_sorted = np.sort(inhibitory_indexes)

ee_from_to = from_to[ np.where( (types[from_to[:,0]]==1) & (types[from_to[:,1]]==1) )[0],:]
ei_from_to = from_to[ np.where( (types[from_to[:,0]]==1) & (types[from_to[:,1]]==0) )[0],:]
ie_from_to = from_to[ np.where( (types[from_to[:,0]]==0) & (types[from_to[:,1]]==1) )[0],:]
ii_from_to = from_to[ np.where( (types[from_to[:,0]]==0) & (types[from_to[:,1]]==0) )[0],:]

ee_from = ee_from_to[:,0]
ee_to = ee_from_to[:,1]
ei_from = ei_from_to[:,0]
ei_to = ei_from_to[:,1]
ie_from = ie_from_to[:,0]
ie_to = ie_from_to[:,1]
ii_from = ii_from_to[:,0]
ii_to = ii_from_to[:,1]

ee_from_idx = npi.indices(excitatory_indexes_sorted, ee_from)
ee_to_idx = npi.indices(excitatory_indexes_sorted, ee_to)
ei_from_idx = npi.indices(excitatory_indexes_sorted, ei_from)
ei_to_idx = npi.indices(inhibitory_indexes_sorted, ei_to)
ie_from_idx = npi.indices(inhibitory_indexes_sorted, ie_from)
ie_to_idx = npi.indices(excitatory_indexes_sorted, ie_to)
ii_from_idx = npi.indices(inhibitory_indexes_sorted, ii_from)
ii_to_idx = npi.indices(inhibitory_indexes_sorted, ii_to)

########################################## CONNECT EXTERNAL STIMULUS

if parameters[p.EXTERNAL_INPUT]:
    stimulus_size = parameters[p.STIMULUS_SIZE]
    stimulus_input_from_to = np.column_stack( (np.arange(stimulus_size),np.random.choice(excitatory_indexes,size = stimulus_size,replace = False)) )
    w_external_exc = parameters[p.WEIGHT_EXTERNAL_INPUT]
else:
    stimulus_size = 0
    stimulus_input_from_to = np.empty([0,2])
    w_external_exc = 0

stimulus_input_from = stimulus_input_from_to[:,0]
stimulus_input_to = stimulus_input_from_to[:,1]
stimulus_input_from_idx = stimulus_input_from
stimulus_input_to_idx = npi.indices(excitatory_indexes_sorted,stimulus_input_to)

########################################## LOAD MODEL PARAMETERS

background_p_exc=parameters[p.BACKGROUND_P_EXC]
background_p_inh=parameters[p.BACKGROUND_P_INH]
taup_exc=parameters[p.EXC_DECAY]*ms
taup_inh=parameters[p.INH_DECAY]*ms
taui_exc=parameters[p.INPUT_DECAY_EX]*ms
taui_inh=parameters[p.INPUT_DECAY_IN]*ms
P_reset_exc=parameters[p.P_RESET_EXC]
P_reset_inh=parameters[p.P_RESET_INH]

tauI_e=parameters[p.INPUT_DECAY_EX]*ms
tauI_i=parameters[p.INPUT_DECAY_IN]*ms

# standalone mode means the model code is compiled in c for faster execution
if configuration[conf.IS_BRIAN_STANDALONE]:
    set_device('cpp_standalone',directory=tmp_dir_path+'output')
    prefs.codegen.cpp.headers += ['"run.h"'] # This is necessary to use brian_end()

# run simulation at a timestep of 1 millisecond
defaultclock.dt = 1*ms
tau_s=1*ms
tau_clock_mult = defaultclock.dt/ms

########################################## NEURON MODEL
eqs = '''
dIe/dt = (-Ie)/tauI_e  : 1
dIi/dt = (-Ii)/tauI_i  : 1
dp/dt = (Ii+Ie)/tau_s + (background_p - p-Ii-Ie) /taup : 1
background_p : 1
reset_p : 1
taup : second
'''

# define two neuron groups comprised of the excitatory and inhibitory poopulations
# the advantage of separating the two is that it allows recording 
neurongroup_exc = NeuronGroup(num_excitatory, eqs, threshold='rand()<=p', reset='p = reset_p',refractory=0*ms,method='euler')
neurongroup_exc.p=background_p_exc
neurongroup_exc.background_p= background_p_exc
neurongroup_exc.reset_p=P_reset_exc
neurongroup_exc.taup=taup_exc

neurongroup_inh = NeuronGroup(num_inhibitory, eqs, threshold='rand()<=p', reset='p = reset_p',refractory=0*ms,method='euler')
neurongroup_inh.p=background_p_inh
neurongroup_inh.background_p= background_p_inh
neurongroup_inh.reset_p=P_reset_inh
neurongroup_inh.taup=taup_inh

if parameters[p.EXTERNAL_INPUT]==True and parameters[p.STIMULUS_SIZE]>0:
    #set up network external input - this assumes default clock at millisecond
    external_input_times=np.zeros(parameters[p.SIMULATION_LENGTH]-1)
    
    #this is assumes clock time of 1ms
    for i in range(len(external_input_times)):
        external_input_times[i]=int(round(750+rand()*500+1000*i))
    
    #define external
    neurongroup_stim = SpikeGeneratorGroup(stimulus_size, \
                                           np.repeat(np.arange(stimulus_size),len(external_input_times)), \
                                           np.tile(external_input_times,stimulus_size)*ms)

net=Network()
net.add(neurongroup_exc)
net.add(neurongroup_inh)

if parameters[p.EXTERNAL_INPUT]==True and parameters[p.STIMULUS_SIZE]>0:  
    net.add(neurongroup_stim)

ee_gain = parameters[p.EE_GAIN]
ei_gain = parameters[p.EI_GAIN]
ie_gain = parameters[p.IE_GAIN]
ii_gain = parameters[p.II_GAIN]

########################################## INITIALIZE SYNAPSES
synapse_model = 'w :1'
pre_model_ee = 'Ie+=ee_gain*w'
pre_model_ei = 'Ie+=ei_gain*w'
pre_model_ie = 'Ii+=ie_gain*w'
pre_model_ii = 'Ii+=ii_gain*w'

Cee = Synapses(neurongroup_exc, neurongroup_exc, synapse_model, on_pre=pre_model_ee,method='euler')
Cei = Synapses(neurongroup_exc, neurongroup_inh, synapse_model, on_pre=pre_model_ei,method='euler')
Cie = Synapses(neurongroup_inh, neurongroup_exc, synapse_model, on_pre=pre_model_ie,method='euler')
Cii = Synapses(neurongroup_inh, neurongroup_inh, synapse_model, on_pre=pre_model_ii,method='euler')

if parameters[p.EXTERNAL_INPUT]==True and parameters[p.STIMULUS_SIZE]>0:
    neuron_model = 'w :1'
    pre_model_external_input = 'Ie+=w'
    Cse = Synapses(neurongroup_stim, neurongroup_exc,neuron_model, on_pre=pre_model_external_input,method='euler')

Cee.connect(i=ee_from_idx, j=ee_to_idx)
Cei.connect(i=ei_from_idx, j=ei_to_idx)
Cie.connect(i=ie_from_idx, j=ie_to_idx)
Cii.connect(i=ii_from_idx, j=ii_to_idx)

Cee.w = wEE
Cei.w = wEI
Cie.w = wIE
Cii.w = wII

net.add(Cee)
net.add(Cei)
net.add(Cie)
net.add(Cii)

if parameters[p.EXTERNAL_INPUT]==True and parameters[p.STIMULUS_SIZE]>0:
    Cse.connect(i=stimulus_input_from_idx,j=stimulus_input_to_idx)
    Cse.w = w_external_exc
    net.add(Cse)

########################################## ADD MONITORS TO RECORD DATA

#monitor the firing rates of the excitatory and inhibitory populations
population_monitor_exc = PopulationRateMonitor(neurongroup_exc)
net.add(population_monitor_exc)

population_monitor_inh = PopulationRateMonitor(neurongroup_inh)
net.add(population_monitor_inh)

#this allows monitoring all individual spikes
if parameters[p.OUTPUT_ACTIVITY]:
    s_mon_exc = SpikeMonitor(neurongroup_exc)
    s_mon_inh = SpikeMonitor(neurongroup_inh)
    net.add(s_mon_exc)
    net.add(s_mon_inh)

########################################## RUN SIMULATION

simduration = parameters[p.SIMULATION_LENGTH] * second
net.run(simduration,report='stdout')

########################################## OUTPUT, ANALYSE

simulation_variables = {'real_path': output_dir_path,
                        'from_to' : from_to, 'types' : types, 'weights' : weights,
                        'population_monitor_exc':population_monitor_exc,
                        'population_monitor_inh':population_monitor_inh}

if parameters[p.OUTPUT_ACTIVITY]:
    simulation_variables['s_mon_exc']=s_mon_exc
    simulation_variables['s_mon_inh']=s_mon_inh
    
if parameters[p.EXTERNAL_INPUT]==True and parameters[p.STIMULUS_SIZE]>0:
    simulation_variables['external_input_times']=external_input_times
else:
    simulation_variables['external_input_times']=[]

output_utils.output_simulation_data(parameters,simulation_variables)

#run analysis after outputting from here
analysis_utils.run_matlab_analysis(output_dir_path)

########################################## CLEANUP
#delete generated code directory
if parameters[p.STANDALONE_MODE]:
    if os.path.isdir(tmp_dir_path):
        shutil.rmtree(tmp_dir_path)
