#     Originally created by Arthur-Ervin Avramiea (2020), arthur.avramiea@gmail.com

#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
# 
#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU General Public License for more details.
# 
#     You should have received a copy of the GNU General Public License
#     along with this program.  If not, see <http://www.gnu.org/licenses/>.

import math
import numpy as np
from . import parameter_utils as p
from scipy.optimize  import minimize
import h5py
from os.path import join
import sys

#for python2, code is faster when using the xrange function
crt_python_version = sys.version_info
if crt_python_version==2:
    range = xrange

#C is the normalization constant to turn the matrix with the exponential of distances e^-distance into a matrix of connectivity probabilities
#distances_exponential is the matrix of the exponential of distance e^-distance
#locRange is the local range
#conP is the connectivity probability in the local range
#total cells is the total number of cells to connect to in the local range
def get_sum_probability_circle(C,distances_exponential,loc_range,con_p,total_cells):
    '''
    Sums all values in the grid normalized by C.
    Returns how far this sum is from the expected value (conProb*total number).
    '''
    connected_on_average = 0

    for i in range(-loc_range,loc_range+1):
        for j in range(-loc_range,loc_range+1):
            connection_probability = distances_exponential[i+loc_range][j+loc_range]*C
            
            if math.pow( i*i+j*j,0.5 )>loc_range:
                continue
            #this is a probability so if higher than 1 reset to 1
            if connection_probability>1:
                connected_on_average+=1
            else:
                connected_on_average+=connection_probability

    return abs(connected_on_average - con_p*total_cells)

#computes the normalization constant C which when multiplied by the inverse exponential distance function, 
#gives the average connectivity probability con_p within local range loc_range
def calc_C_exponential_circle(con_p,loc_range):
    grid = np.zeros((loc_range*2+1,loc_range*2+1))
    total_cells = 0

    for i in range(-loc_range,loc_range+1):
        for j in range(-loc_range,loc_range+1):
            grid[i+loc_range][j+loc_range]=math.exp(-math.pow(i*i+j*j,0.5))
            if math.pow( i*i+j*j,0.5 )>loc_range or (i==0 and j==0):
                grid[i+loc_range][j+loc_range]=0
            else:
                total_cells+=1
    
    optimization_result = minimize(get_sum_probability_circle,0.1,args=(grid,loc_range,con_p,total_cells),method='nelder-mead',options={'xtol':1,'disp':False})
    return optimization_result.x[0]

##Used for getting coordinates.
##grid_points_to_sample==num_inhibitory, grid_width = network size (number of neurons)
def get_random_points_placement(grid_width,grid_points_to_sample):
    '''
    int grid_width, grid_points_to_sample--> arr[np.array,np.array]
    Returns a list of indices of neurons which would best suited to be inhibitory according to
    Mitchells approach
    '''

    placed_points = np.zeros((grid_points_to_sample,2),dtype=int)
    ##Make neighbor samples also inputtable?
    neighbor_samples = 50
    
    grid_index_choice = np.arange(grid_width)
    
    #Find the most minimum distanced places in the grid among those placed and randomly chosen.
    ##In the beginning minimum is diagonal distance.
    ##among randomly picked samples
    for placed_point_idx in range(grid_points_to_sample):
        random_generated_dots = np.random.choice(grid_index_choice,size=(neighbor_samples,2),replace=True)
        generated_dots_minimum_distance = np.zeros(neighbor_samples)
        #generate best dot
        for candidate_idx in range(neighbor_samples):
            minimum_distance = grid_width*np.sqrt(2)
            if placed_point_idx>0:
                all_dots_distance = np.sqrt( ( random_generated_dots[candidate_idx,0] - placed_points[:,0])**2  + ( random_generated_dots[candidate_idx,1] - placed_points[:,1])**2 )
                minimum_distance = np.min(all_dots_distance)
            generated_dots_minimum_distance[candidate_idx]=minimum_distance
            
        best_candidate_idx = np.argmax(generated_dots_minimum_distance)
        best_candidate_distance = generated_dots_minimum_distance[best_candidate_idx]
        
        assert best_candidate_distance>0
        
        placed_points[placed_point_idx,0]=random_generated_dots[best_candidate_idx,0]
        placed_points[placed_point_idx,1]=random_generated_dots[best_candidate_idx,1]
    
    xx = placed_points[:,0]
    yy = placed_points[:,1]
        
    return [xx,yy]

#get neuron coordinates based on index
def get_coordinates(i,grid_size):
    row = i // grid_size
    col = i % grid_size
    return [row,col]

def get_neuron_idx(row,col,grid_size):
    neuron_idx = row*grid_size+col
    return neuron_idx
    
def get_network(num_neurons,types,parameters):
    from_to=get_network_circle_neighborhood_exponential(num_neurons,types,parameters)
     
    return from_to

# implementation from https://gist.github.com/lj8385174/979cb843f0681173b7cea768d53432d5
def cartesian(arrays, out=None):
    """
    Generate a cartesian product of input arrays.

    Parameters
    ----------
    arrays : list of array-like
        1-D arrays to form the cartesian product of.
    out : ndarray
        Array to place the cartesian product in.

    Returns
    -------
    out : ndarray
        2-D array of shape (M, len(arrays)) containing cartesian products
        formed of input arrays.

    Examples
    --------
    >>> cartesian(([1, 2, 3], [4, 5], [6, 7]))
    array([[1, 4, 6],
           [1, 4, 7],
           [1, 5, 6],
           [1, 5, 7],
           [2, 4, 6],
           [2, 4, 7],
           [2, 5, 6],
           [2, 5, 7],
           [3, 4, 6],
           [3, 4, 7],
           [3, 5, 6],
           [3, 5, 7]])

    """

    arrays = [np.asarray(x) for x in arrays]
    dtype = arrays[0].dtype

    n = np.prod([x.size for x in arrays])
    if out is None:
        out = np.zeros([n, len(arrays)], dtype=dtype)

    m = int(n / arrays[0].size)
    out[:,0] = np.repeat(arrays[0], m)
    if arrays[1:]:
        cartesian(arrays[1:], out=out[0:m,1:])
        for j in range(1, arrays[0].size):
            out[j*m:(j+1)*m,1:] = out[0:m,1:]
    return out

#select neighbors to connect to according to the probabillistic connectivity rule. P(ij) = C*e^-(euclidean_distance(ij))
def get_neighbor_neurons_exp(all_neighborhood_distances,all_neighborhood_indexes,normalization_constant):
    #determine euclidean distance to all neurons in the neighborhood
    euclidean_distance = np.sqrt(np.power(all_neighborhood_distances[:,0],2) + np.power(all_neighborhood_distances[:,1],2))
    
    #select neighbors to connect to according to the probabillistic connectivity rule. P(ij) = C*e^-(euclidean_distance(ij))
    random_values = np.random.rand(np.shape(all_neighborhood_distances)[0])
    selected_outgoing_indexes = np.where(random_values<normalization_constant*np.exp(- euclidean_distance ))

    if(len(selected_outgoing_indexes[0])>0):
        selected_outgoing_actual_neuron_indexes = (all_neighborhood_indexes)[selected_outgoing_indexes]
    else:
        selected_outgoing_actual_neuron_indexes=[]

    return selected_outgoing_actual_neuron_indexes

def get_network_circle_neighborhood_exponential(num_neurons,types,parameters):
    #load relevant parameters 
    grid_size=int(parameters[p.GRID_SIZE])
    neuron_range_exc=int(parameters[p.LOCAL_RANGE_EX])
    neuron_range_inh=int(parameters[p.LOCAL_RANGE_IN])
    connectivity_exc = parameters[p.EXCITATORY_CONNECTIVITY]
    connectivity_inh = parameters[p.INHIBITORY_CONNECTIVITY]

    from_to = []

    #compute the normalization constant that turns distances into probabilities
    exc_C = calc_C_exponential_circle(connectivity_exc,neuron_range_exc)
    inh_C = calc_C_exponential_circle(connectivity_inh,neuron_range_inh)
    connectivity_function = get_neighbor_neurons_exp

    for i in range(num_neurons):
        [from_row,from_col]=get_coordinates(i,grid_size)
        if types[i]==p.TYPE_EXC:
            crt_range = neuron_range_exc
            normalization_C = exc_C
        else:
            crt_range = neuron_range_inh
            normalization_C = inh_C
        
        ### first we consider all neighbors within a square neighborhood
         
        # contains on first column row distance, on second column column distance
        square_neighborhood_all_distances_relative_to_neuron = cartesian([np.arange(-crt_range,crt_range+1),np.arange(-crt_range,crt_range+1)])
        square_neighborhood_all_euclidean_distances_relative_to_neuron = np.sqrt(np.power(square_neighborhood_all_distances_relative_to_neuron[:,0],2) +
                                                                              np.power(square_neighborhood_all_distances_relative_to_neuron[:,1],2))
        circle_neighborhood_all_distances_relative_to_neuron = square_neighborhood_all_distances_relative_to_neuron[np.where(square_neighborhood_all_euclidean_distances_relative_to_neuron<=crt_range)[0],:]        
        
        #get rid of current neuron (no synapses of a neuron to itself)
        index_of_current_neuron = np.where(np.logical_and(circle_neighborhood_all_distances_relative_to_neuron[:,0]==0 , circle_neighborhood_all_distances_relative_to_neuron[:,1]==0))
        get_rid_of_indexes = index_of_current_neuron[0]
        circle_neighborhood_all_distances_relative_to_neuron = circle_neighborhood_all_distances_relative_to_neuron[np.setdiff1d( np.arange(len(circle_neighborhood_all_distances_relative_to_neuron)), get_rid_of_indexes ),:]

        # computes absolute index of neighbor rows and neighbor columns
        neighborhood_rows = from_row + circle_neighborhood_all_distances_relative_to_neuron[:,0]
        neighborhood_cols = from_col + circle_neighborhood_all_distances_relative_to_neuron[:,1]
        
        ### next we decide what to do with neighbors that fall of the edges of the grid
        #if torus wrap around is true, this allows neurons on the edge are neighbors with neighbors on the opposite edge
        if parameters[p.TORUS_WRAP_AROUND]:
            outside_left = np.where(neighborhood_cols<0)
            neighborhood_cols[outside_left]=grid_size+neighborhood_cols[outside_left]
            outside_right = np.where(neighborhood_cols>=grid_size)
            neighborhood_cols[outside_right]=neighborhood_cols[outside_right]-grid_size
            
            outside_top = np.where(neighborhood_rows<0)
            neighborhood_rows[outside_top]=grid_size+neighborhood_rows[outside_top]
            outside_bottom = np.where(neighborhood_rows>=grid_size)
            neighborhood_rows[outside_bottom]=neighborhood_rows[outside_bottom]-grid_size
            
            legal_neighborhood_indexes_within_neighborhood=np.arange(len(neighborhood_rows))
            legal_neighborhood_distances = circle_neighborhood_all_distances_relative_to_neuron[legal_neighborhood_indexes_within_neighborhood,:]
        #if we do not wrap around the edges, we remove the neighbors that fall beyond the edges
        else:
            legal_neighborhood_indexes_within_neighborhood=np.logical_and.reduce(np.array( (neighborhood_rows>=0 , neighborhood_cols>=0, neighborhood_rows<grid_size, neighborhood_cols<grid_size)))
            legal_neighborhood_distances = circle_neighborhood_all_distances_relative_to_neuron[legal_neighborhood_indexes_within_neighborhood,:]
        
        ### the remaining are the neighbors that we can connect to
        legal_neighborhood_rows = neighborhood_rows[legal_neighborhood_indexes_within_neighborhood]
        legal_neighborhood_cols = neighborhood_cols[legal_neighborhood_indexes_within_neighborhood]
        legal_neighborhood_indexes = legal_neighborhood_rows * grid_size + legal_neighborhood_cols
        
        #this gives the subset of neurons within the neighborhood that are excitatory and inhibitory
        all_outgoing_exc_indexes_within_neighborhood = np.where(types[legal_neighborhood_indexes]==p.TYPE_EXC)
        all_outgoing_inh_indexes_within_neighborhood = np.where(types[legal_neighborhood_indexes]==p.TYPE_INH)
        
        all_outgoing_exc_distances = legal_neighborhood_distances[all_outgoing_exc_indexes_within_neighborhood]
        all_outgoing_exc_absolute_indexes = legal_neighborhood_indexes[all_outgoing_exc_indexes_within_neighborhood]
        all_outgoing_inh_distances = legal_neighborhood_distances[all_outgoing_inh_indexes_within_neighborhood]
        all_outgoing_inh_absolute_indexes = legal_neighborhood_indexes[all_outgoing_inh_indexes_within_neighborhood]
        
        #then we use the connectivity function which takes the vector of all neighbors and gives back those that we connect to, given the connectivity function  
        selected_outgoing_exc_indexes=connectivity_function(all_outgoing_exc_distances,all_outgoing_exc_absolute_indexes,normalization_C)
        if len(selected_outgoing_exc_indexes)>0:
            for j in range(len(selected_outgoing_exc_indexes)):
                from_to.append([i,selected_outgoing_exc_indexes[j]])

        selected_outgoing_inh_indexes=connectivity_function(all_outgoing_inh_distances,all_outgoing_inh_absolute_indexes,normalization_C)
        if len(selected_outgoing_inh_indexes)>0:
            for j in range(len(selected_outgoing_inh_indexes)):
                from_to.append([i,selected_outgoing_inh_indexes[j]])

    return np.array(from_to)

def get_network_arrays_from_path(path):
    f = h5py.File(path, "r")
    
    types = np.array(f["networks/main_network/types"])
    main_network_from_to=np.array(f["connections/from_main_to_main/indexes"])
    main_network_weights=np.array(f["connections/from_main_to_main/weights"])
    
    f.close()
    return [ types,main_network_from_to,main_network_weights ]

def write_network_structure(path,types,from_to,weights):
    f = h5py.File(join(path,"SimulationData.h5"), "a")
    
    f.create_dataset("networks/main_network/types",types.shape,dtype="i",data=types, compression="gzip", compression_opts=9)
    main_network_group = f["networks/main_network"]
    main_network_group.attrs["index_network"]=0
    main_network_group.attrs["type"]="grid_network"
    
    f.create_dataset("connections/from_main_to_main/indexes", from_to.shape, dtype='i', data=from_to, compression="gzip", compression_opts=9)
    f.create_dataset("connections/from_main_to_main/weights", weights.shape, dtype='f', data=weights, compression="gzip", compression_opts=9)
    from_main_to_main_connections = f["connections/from_main_to_main"]
    from_main_to_main_connections.attrs["from_network_index"]=0
    from_main_to_main_connections.attrs["to_network_index"]=0
    
    f.close()