#     Originally created by Arthur-Ervin Avramiea (2020), arthur.avramiea@gmail.com

#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
# 
#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU General Public License for more details.
# 
#     You should have received a copy of the GNU General Public License
#     along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import h5py

from brian2 import *

from . import parameter_utils as p
from configuration import configuration as conf
from . import network_utils

def get_output_directory_name(parameters):    
    output_dir_param_keys = parameters[p.OUTPUT_DIR_PATH_PARAMETERS].split(",")
    output_dir_param_values = []
    for parameter_key in output_dir_param_keys:
        output_dir_param_values.append(parameters[parameter_key.strip()])
    output_directory_name = parameters[p.OUTPUT_DIR_PATH_FORMAT] % tuple(output_dir_param_values)

    return output_directory_name

def get_output_dir_path(param):
    configuration = conf.get_configuration()
    data_output_path = configuration[conf.DATA_FILE_PATH]
    
    return data_output_path + get_output_directory_name(param)

def remove_file_if_exists(path_to_file):
    if os.path.exists(path_to_file):
        os.remove(path_to_file)

def create_simulation_directories(params):
    configuration = conf.get_configuration()
    data_output_path = configuration[conf.DATA_FILE_PATH]
    tmp_output_path = configuration[conf.TMP_FILE_PATH]
    
    output_directory_name = get_output_directory_name(params)
            
    output_dir_path = data_output_path + output_directory_name
    tmp_dir_path = tmp_output_path + output_directory_name

    if not os.path.exists(output_dir_path):
        os.makedirs(output_dir_path)
    
    if not os.path.exists(tmp_dir_path):
        os.makedirs(tmp_dir_path)
        
    with open(output_dir_path+"Simulation.Parameters","w") as f:
        for crt_key in sorted(params.keys()):
            crt_value = params[crt_key]
            f.write(crt_key+"="+str(crt_value)+"\n")
        
    return [output_dir_path,tmp_dir_path]
  
def output_simulation_data(parameters,simulation_variables):
    output_dir_path = simulation_variables['real_path']
    
    population_monitor_exc = simulation_variables['population_monitor_exc']
    population_monitor_inh = simulation_variables['population_monitor_inh']
    types = simulation_variables['types']
    from_to = simulation_variables['from_to']
    weights = simulation_variables['weights']
    Fs = parameters[p.SAMPLING_FREQUENCY]
    
    external_input_times = simulation_variables['external_input_times']
    
    #get counts for each synapse type
    ee_count = len( np.where( (types[from_to[:,0]]==1)  & (types[from_to[:,1]]==1) )[0] )
    ei_count = len( np.where( (types[from_to[:,0]]==1) & (types[from_to[:,1]]==0) )[0] )
    ie_count = len( np.where( (types[from_to[:,0]]==0) & (types[from_to[:,1]]==1) )[0] )
    ii_count = len( np.where( (types[from_to[:,0]]==0) & (types[from_to[:,1]]==0) )[0] )
    
    #get neuron coutns for each type
    num_excitatory = sum(types==1)
    num_inhibitory = sum(types==0)
    
    #turn from rates to synapse counts
    spike_ms_times_exc = (population_monitor_exc.rate/Hz*num_excitatory*1/Fs)
    spike_ms_times_inh = (population_monitor_inh.rate/Hz*num_inhibitory*1/Fs)
    
    #first create new empty simulation data hdf5 file (overwrites old one if existing)
    #and then append to it
    f = h5py.File(output_dir_path+"SimulationData.h5", "w")
    f.close()
    f = h5py.File(output_dir_path+"SimulationData.h5", "a")
    
    f.create_dataset("data/main_network_excitatory_spikes_timeseries",spike_ms_times_exc.shape,dtype="f",\
                     data=spike_ms_times_exc, compression="gzip", compression_opts=9)
    f.create_dataset("data/main_network_inhibitory_spikes_timeseries",spike_ms_times_inh.shape,dtype="f",\
                     data=spike_ms_times_inh, compression="gzip", compression_opts=9)
    if len(external_input_times)>0:
        f.create_dataset("data/external_input_spike_times",external_input_times.shape,dtype="f",\
                         data=external_input_times, compression="gzip", compression_opts=9)
    f.close()
    
    #overwrite file if exists
    with open(output_dir_path+"NetworkStatistics.txt","w") as f:
        f.write("eeCount="+str(ee_count)+"\n"+\
                "eiCount="+str(ei_count)+"\n"+\
                "ieCount="+str(ie_count)+"\n"+\
                "iiCount="+str(ii_count))
    
    #overwrite file if exists
    with open(output_dir_path+"ActivityStatistics.txt","w") as f:
        f.write("meanFiringrateExc="+str(np.mean(population_monitor_exc.rate/Hz))+"\n"+\
                "meanFiringrateInh="+str(np.mean(population_monitor_inh.rate/Hz))+"\n")
            
    if parameters[p.OUTPUT_ACTIVITY]:
        s_mon_exc = simulation_variables['s_mon_exc']
        s_mon_inh = simulation_variables['s_mon_inh']
        
        f = h5py.File(output_dir_path+"SimulationData.h5", "a")
        
        f.create_dataset("data/main_network_excitatory_spikes_individual_neurons",s_mon_exc.it[0].shape,dtype="i",\
                     data=s_mon_exc.it[0], compression="gzip", compression_opts=9)
        f.create_dataset("data/main_network_excitatory_spikes_individual_times",s_mon_exc.t.shape,dtype="i",\
                     data=s_mon_exc.t/ms, compression="gzip", compression_opts=9)
                     
        f.create_dataset("data/main_network_inhibitory_spikes_individual_neurons",s_mon_inh.it[0].shape,dtype="i",\
                     data=s_mon_inh.it[0], compression="gzip", compression_opts=9)
        f.create_dataset("data/main_network_inhibitory_spikes_individual_times",s_mon_inh.t.shape,dtype="i",\
                     data=s_mon_inh.t/ms, compression="gzip", compression_opts=9)

    if parameters[p.OUTPUT_NETWORK]:
        network_utils.write_network_structure(output_dir_path,types,from_to,weights)